/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package org.apache.wsif.compiler.util;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.net.MalformedURLException;

import org.apache.wsif.WSIFException;
import org.apache.wsif.compiler.schema.tools.Conventions;

/**
 * @author Matthew J. Duftler
 */
public class StreamFactory {
	
    public OutputStream getOutputStream(
        String root,
        String name,
        boolean overwrite)
        throws WSIFException {
        boolean verbose = Conventions.getVerbose();

        if (root != null) {
            File directory = new File(root);

            if (!directory.exists()) {
                if (!directory.mkdirs()) {
                    throw new WSIFException("Failed to create directory '" + root + "'.");
                } else if (verbose) {
                    System.out.println("Created directory '" + directory.getAbsolutePath() + "'.");
                }
            }
        }

        File file = new File(root, name);
        String absolutePath = file.getAbsolutePath();

        if (file.exists()) {
            if (!overwrite) {
                throw new WSIFException(
                    "File '"
                        + absolutePath
                        + "' already exists. Please remove it or "
                        + "enable the overwrite option.");
            } else {
                file.delete();

                if (verbose) {
                    System.out.println("Deleted file '" + absolutePath + "'.");
                }
            }
        }

        if (verbose) {
            System.out.println("Created file '" + absolutePath + "'.");
        }

        try {
            return new FileOutputStream(absolutePath);
        } catch (FileNotFoundException e) {
            throw new WSIFException("Problem getting output stream.", e);
        }
    }

    public InputStream getInputStream(String root, String name)
        throws IOException {
        String fileName = (root != null) ? root + File.separatorChar + name : name;
        URL url = null;
        Object content = null;

        try {
            url = getURL(null, fileName, 1);
            content = url.getContent();
        } catch (SecurityException e) {
            throw new IOException(
                "Your JVM's security manager has disallowed "
                    + "access to '"
                    + fileName
                    + "'.");
        } catch (IOException e) {
            throw new IOException("The resource at '" + fileName + "' was not found.");
        }

        if (content == null) {
            throw new IllegalArgumentException("No content at '" + fileName + "'.");
        } else if (content instanceof InputStream) {
            return (InputStream) content;
        } else {
            throw new IOException("The content of '" + fileName + "' is not a stream.");
        }
    }

    private static URL getURL(URL contextURL, String spec, int recursiveDepth)
        throws MalformedURLException {
        URL url = null;
    
        try {
            url = new URL(contextURL, spec);
    
            try {
                url.openStream();
            } catch (IOException ioe1) {
                throw new MalformedURLException("This file was not found: " + url);
            }
        } catch (MalformedURLException e1) {
            url = new URL("file", "", spec);
    
            try {
                url.openStream();
            } catch (IOException ioe2) {
                if (contextURL != null) {
                    String contextFileName = contextURL.getFile();
                    String parentName = new File(contextFileName).getParent();
    
                    if (parentName != null && recursiveDepth < 3) {
                        return getURL(
                            new URL("file", "", parentName + '/'),
                            spec,
                            recursiveDepth + 1);
                    }
                }
    
                throw new MalformedURLException("This file was not found: " + url);
            }
        }
    
        return url;
    }
}