/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package org.apache.wsif.compiler.util;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.io.StringWriter;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.wsdl.Definition;
import javax.wsdl.Import;
import javax.wsdl.Types;
import javax.wsdl.extensions.UnknownExtensibilityElement;
import javax.xml.namespace.QName;

import org.apache.wsif.WSIFConstants;
import org.apache.wsif.WSIFException;

import com.ibm.wsdl.util.StringUtils;

/**
 * @author Matthew J. Duftler
 */
public class Utils {
	
    public static void addAllTypesElements(Definition def, List toList) {
        Types types = def.getTypes();
        if (types != null) {
            Iterator extEleIt = types.getExtensibilityElements().iterator();
            while (extEleIt.hasNext()) {
                // the unknown extensibility element are wrappers for DOM elements
                UnknownExtensibilityElement typesElement =
                    (UnknownExtensibilityElement) extEleIt.next();
                toList.add(typesElement);
            }
        }

        Map imports = def.getImports();

        if (imports != null) {
            Iterator valueIterator = imports.values().iterator();

            while (valueIterator.hasNext()) {
                List importList = (List) valueIterator.next();

                if (importList != null) {
                    Iterator importIterator = importList.iterator();

                    while (importIterator.hasNext()) {
                        Import tempImport = (Import) importIterator.next();

                        if (tempImport != null) {
                            Definition importedDef = tempImport.getDefinition();

                            if (importedDef != null) {
                                addAllTypesElements(importedDef, toList);
                            }
                        }
                    }
                }
            }
        }
    }

    public static List getAllTypesElements(Definition def) {
        List ret = new Vector();

        addAllTypesElements(def, ret);

        return ret;
    }

    public static String getPackageName(String fqClassName) {
        String packageName = "";

        if (fqClassName != null) {
            int index = fqClassName.lastIndexOf('.');

            if (index != -1) {
                packageName = fqClassName.substring(0, index);
            }
        }

        return packageName;
    }

    public static String getClassName(String fqClassName) {
        if (fqClassName != null) {
            int index = fqClassName.lastIndexOf('.');

            if (index != -1) {
                fqClassName = fqClassName.substring(index + 1);
            }
        }

        return fqClassName;
    }

    public static String queryJavaTypeName(
        QName type,
        String encodingStyleURI,
        Hashtable typeReg)
        throws IllegalArgumentException {
        TypeMapping tm = (TypeMapping) typeReg.get(type);

        if (tm != null) {
            return tm.javaType;
        } else if (
            encodingStyleURI != null
                && encodingStyleURI.equals(WSIFConstants.NS_URI_LITERAL_XML)) {
            return "org.w3c.dom.Element";
        } else {
            throw new IllegalArgumentException("No mapping was found for '" + type + "'.");
        }
    }

    public static String getQuotedString(Reader source, int indent)
        throws WSIFException {
        String indentStr = StringUtils.getChars(indent, ' ');

        try {
            BufferedReader br = new BufferedReader(source);
            StringWriter sw = new StringWriter();
            int count = 0;
            String tempLine = null;

            while ((tempLine = br.readLine()) != null) {
                sw.write(
                    (count > 0
                        ? " + \"" + StringUtils.lineSeparatorStr + "\" +" + StringUtils.lineSeparator
                        : "")
                        + indentStr
                        + '\"'
                        + StringUtils.cleanString(tempLine)
                        + '\"');

                count++;
            }

            return sw.toString();
        } catch (IOException e) {
            throw new WSIFException("Problem writing strings.", e);
        }
    }

    public static String convertToObject(String sourceClassName, String expr)
        throws WSIFException {
        return convertClass(sourceClassName, expr, "java.lang.Object");
    }

    public static String convertFromObject(String expr, String targetClassName)
        throws WSIFException {
        return convertClass("java.lang.Object", expr, targetClassName);
    }

    private static String convertClass(
        String sourceClassName,
        String expr,
        String targetClassName)
        throws WSIFException {
        if (sourceClassName == null || targetClassName == null) {
            throw new WSIFException(
                "I was unable to convert an object from "
                    + sourceClassName
                    + " to "
                    + targetClassName
                    + ".");
        }

        String shortTargetClassName = getShortName(targetClassName);

        if (sourceClassName.equals("java.lang.Object")) {
            if (isPrimitive(targetClassName)) {
                return "(("
                    + getWrapperClassName(targetClassName)
                    + ")"
                    + expr
                    + ")."
                    + shortTargetClassName
                    + "Value()";
            } else {
                return "(" + targetClassName + ")" + expr;
            }
        } else if (
            isPrimitive(sourceClassName) && targetClassName.equals("java.lang.Object")) {
            return "new " + getWrapperClassName(sourceClassName) + "(" + expr + ")";
        } else {
            // Target class must be "assignableFrom" source class.
            return expr;
        }
    }

    private static String getShortName(String className) {
        if (className.startsWith("java.lang."))
            return className.substring(10);
        else
            return className;
    }

    private static String getWrapperClassName(String primitiveClassName) {
        if (primitiveClassName.equals("int"))
            return "Integer";
        else if (primitiveClassName.equals("char"))
            return "Character";
        else
            return getCapitalized(primitiveClassName);
    }

    private static String getCapitalized(String className) {
        return Character.toUpperCase(className.charAt(0)) + className.substring(1);
    }

    private static boolean isPrimitive(String className) {
        String[] primNames =
            {
                "boolean",
                "byte",
                "char",
                "short",
                "int",
                "long",
                "float",
                "double",
                "void" };

        for (int i = 0; i < primNames.length; i++) {
            if (primNames[i].equals(className)) {
                return true;
            }
        }
        return false;
    }
}