/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package org.apache.wsif.providers.soap.apachesoap;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Serializable;
import java.io.StringReader;
import java.io.StringWriter;
import java.net.URL;
import java.util.Hashtable;
import java.util.HashMap;

import javax.mail.MessagingException;
import org.apache.soap.Envelope;
import org.apache.soap.SOAPException;
import org.apache.soap.encoding.SOAPMappingRegistry;
import org.apache.soap.rpc.SOAPContext;
import org.apache.soap.transport.SOAPTransport;
import org.apache.wsif.WSIFCorrelationId;
import org.apache.wsif.WSIFCorrelationService;
import org.apache.wsif.WSIFException;
import org.apache.wsif.logging.Trc;
import org.apache.wsif.util.WSIFCorrelationServiceLocator;
import org.apache.wsif.util.WSIFProperties;
import org.apache.wsif.util.jms.WSIFJMSCorrelationId;
import org.apache.wsif.util.jms.WSIFJMSDestination;
import org.apache.wsif.util.jms.WSIFJMSFinder;
import org.apache.wsif.wsdl.extensions.jms.JMSAddress;

/**
 * This class is a SOAPTransport that supports JMS.
 *
 * @author Mark Whitlock <whitlock@apache.org>
 */
public class SOAPJMSConnection implements SOAPTransport {
    private BufferedReader responseReader = null;
    private SOAPContext responseSOAPContext = null;
    private WSIFJMSDestination destination = null;

    // folowing are for async operation
    private boolean asyncOperation = false;
    private WSIFOperation_ApacheSOAP wsifOperation = null;
    private long syncTimeout;
    private long asyncTimeout;

    private static final String DUMMY_RESPONSE =
        "<?xml version='1.0' encoding='UTF-8'?>\n<SOAP-ENV:Envelope xmlns:SOAP-ENV=\"http://schemas.xmlsoap.org/soap/envelope/\" xmlns:xsi=\"http://www.w3.org/1999/XMLSchema-instance\" xmlns:xsd=\"http://www.w3.org/1999/XMLSchema\">\n<SOAP-ENV:Body>\n<ns1:addEntryResponse xmlns:ns1=\"http://wsifservice.addressbook/\" SOAP-ENV:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\">\n</ns1:addEntryResponse>\n\n</SOAP-ENV:Body>\n</SOAP-ENV:Envelope>";

    public SOAPJMSConnection(JMSAddress ja, String portName) throws WSIFException {
        Trc.entry(this, ja, portName);
        destination =
            new WSIFJMSDestination(
                WSIFJMSFinder.newFinder(ja,portName),
                ja.getJmsProvDestName(),
                WSIFProperties.getSyncTimeout());
        Trc.exit(ja);
    }

    public Hashtable getHeaders() {
        Trc.entry(this);
        Trc.exit(null);
        return null;
    }

    public SOAPContext getResponseSOAPContext() {
        Trc.entry(this);
        Trc.exit(responseSOAPContext);
        return responseSOAPContext;
    }

    public long getSyncTimeout() {
        Trc.entry(this);
        Trc.exit(new Long(syncTimeout));
        return syncTimeout;
    }

    public long getAsyncTimeout() {
        Trc.entry(this);
        Trc.exit(new Long(asyncTimeout));
        return asyncTimeout;
    }

    public BufferedReader receive() {
        Trc.entry(this);
        Trc.exit();
        return responseReader;
    }

    /**
     * This send method is really doing a send followed by a receive. The
     * receive method just returns the BufferedReader that is set up by
     * this method.
     */
    public void send(
        URL sendTo,
        String action,
        Hashtable headers,
        Envelope env,
        SOAPMappingRegistry smr,
        SOAPContext ctx)
        throws SOAPException {
        Trc.entry(this, sendTo, action, headers, env, smr, ctx);

        try {
            if (isAsyncOperation()) {
                performAsyncSend(sendTo, action, headers, env, smr, ctx);
            } else {
                StringWriter payloadSW = new StringWriter();
                env.marshall(payloadSW, smr, ctx);
                String id = destination.send(payloadSW.toString(), null);
                String response = destination.receiveString(id, syncTimeout);
                responseSOAPContext = new SOAPContext();
                responseSOAPContext.setRootPart(response, "text/xml");
                responseReader = new BufferedReader(new StringReader(response));
            }
        } catch (IOException ioe) {
        	Trc.exception(ioe);
            // Not sure what the faultCode should be - this may be wrong
            throw new SOAPException("WSIF SOAPJMSConnection ", ioe.toString());
        } catch (MessagingException me) {
        	Trc.exception(me);
            // Not sure what the faultCode should be - this may be wrong
            throw new SOAPException("WSIF SOAPJMSConnection ", me.toString());
        }
        Trc.exit();
    }

    /**
     * Send the request asynchronously.
     * After sending the request this associates the JMS message correlation
     * ID with the WSIFOperation in the correlation service, and stores the
     * the correlation ID in the message context so the WSIFOperation can
     * pass it back to the executeRequestResponseAsync caller.  
     */
    private void performAsyncSend(
        URL sendTo,
        String action,
        Hashtable headers,
        Envelope env,
        SOAPMappingRegistry smr,
        SOAPContext ctx)
        throws IOException, SOAPException, MessagingException {
        String msgID;
        StringWriter payloadSW = new StringWriter();
        env.marshall(payloadSW, smr, ctx);
        WSIFOperation_ApacheSOAP wsifOp = (WSIFOperation_ApacheSOAP) getWsifOperation();

        WSIFCorrelationId cid;

        // only save op in the correlation service if there's a response handler
        if ( wsifOp.getResponseHandler() == null ) {
            msgID = destination.send(payloadSW.toString(), null);
            cid = new WSIFJMSCorrelationId( msgID );
        } else {
           WSIFCorrelationService correlator =
              WSIFCorrelationServiceLocator.getCorrelationService();
           synchronized( correlator ) {   
              msgID = destination.send(payloadSW.toString(), null);
              cid = new WSIFJMSCorrelationId( msgID );
              if ( correlator != null ) {
                 correlator.put( 
                    cid,
                    (Serializable)getWsifOperation(),
                    asyncTimeout );
              }
           }
        }
           

        wsifOp.setAsyncRequestID(new WSIFJMSCorrelationId(msgID));

        // SOAP doesn't like a null response so give it a dummy null
        responseSOAPContext = new SOAPContext();
        responseSOAPContext.setRootPart(DUMMY_RESPONSE, "text/xml");
        responseReader = new BufferedReader(new StringReader(DUMMY_RESPONSE));

    }

    public void setAsyncOperation(boolean b) {
        Trc.entry(this, b);
        asyncOperation = b;
        Trc.exit();
    }

    public boolean isAsyncOperation() {
        Trc.entry(this);
        Trc.exit(asyncOperation);
        return asyncOperation;
    }

    public WSIFOperation_ApacheSOAP getWsifOperation() {
        Trc.entry(this);
        Trc.exit(wsifOperation);
        return wsifOperation;
    }

    public void setWsifOperation(WSIFOperation_ApacheSOAP op) {
        Trc.entry(this, op);
        wsifOperation = op;
        Trc.exit();
    }

    public void setJmsProperty(String name, Object value) throws WSIFException {
        Trc.entry(this, name, value);
        destination.setProperty(name, value);
        Trc.exit();
    }

    public void setJmsProperties(HashMap hm) { 
        Trc.entry(this,hm);
        destination.setProperties(hm);  
        Trc.exit();
    }

    public HashMap getJmsProperties() throws WSIFException { 
  	   Trc.entry(this);
  	   HashMap hm = destination.getProperties();  
  	   Trc.exit(hm);
  	   return hm;  
    }

    void close() throws WSIFException {
        Trc.entry(this);
        destination.close();
        Trc.exit();
    }

    public void setSyncTimeout(long timeout) {
        Trc.entry(this, new Long(timeout));
        syncTimeout = timeout;
        Trc.exit();
    }

    public void setAsyncTimeout(long timeout) {
        Trc.entry(this, new Long(timeout));
        asyncTimeout = timeout;
        Trc.exit();
    }
}